library ieee;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
library work;
use work.std_logic_arithext.all;


--datapath entity
entity crypto_fpga is
   port(
      init:in std_logic;
      EoM:in std_logic;
      load:in std_logic;
      fetch:in std_logic;
      ack:out std_logic;
      idata:in std_logic_vector(15 downto 0);
      odata:out std_logic_vector(15 downto 0);
      rst_n : in std_logic;
      clk : in std_logic

   );
end crypto_fpga;


--signal declaration
architecture RTL of crypto_fpga is
signal load_reg:std_logic;
signal load_reg_wire:std_logic;
signal fetch_reg:std_logic;
signal fetch_reg_wire:std_logic;
signal init_reg:std_logic;
signal init_reg_wire:std_logic;
signal load_ack:std_logic;
signal fetch_ack:std_logic;
signal s_idata:std_logic_vector(15 downto 0);
signal s_odata:std_logic_vector(15 downto 0);
signal accept_load:std_logic;
signal odata_ready:std_logic;
signal ack_reg:std_logic;
signal ack_reg_wire:std_logic;
signal message_reg:std_logic_vector(511 downto 0);
signal message_reg_wire:std_logic_vector(511 downto 0);
signal r_shift:std_logic;
signal start:std_logic;
signal round:std_logic_vector(4 downto 0);
signal round_wire:std_logic_vector(4 downto 0);
signal sig_0:std_logic;
signal sig_1:std_logic;
signal sig_2:std_logic;
signal sig_3:std_logic;
signal sig_4:std_logic;
signal sig_5:std_logic;
signal sig_6:std_logic_vector(511 downto 0);
signal sig_7:std_logic_vector(4 downto 0);
signal odata_int:std_logic_vector(15 downto 0);
signal ack_int:std_logic;
signal sig_8:std_logic;
signal sig_9:std_logic;
signal sig_10:std_logic;
signal sig_11:std_logic;


--component map declaration
component CORE
   port(
      message:in std_logic_vector(511 downto 0);
      Hashvalue:out std_logic_vector(15 downto 0);
      start:in std_logic;
      EoM:in std_logic;
      shift:out std_logic;
      rst_n : in std_logic;
      clk : in std_logic
   );
end component;
type STATE_TYPE is (S_uninit,S_ready,S_loading,S_calc,S_fetch,S_loadstep1,S_loadstep2,S_fetchstep1,S_fetchstep2);
signal STATE:STATE_TYPE;
type CONTROL is (initializedef
, nothingstart0odata0accept_load0def
, start0accept_load1odata0def
, take_in_odataaccept_load0start0odata0def
, start1accept_load0odata0def
, round0start0accept_load1odata0def
, round1start0accept_load1odata0def
, accept_load0start0odata0def
, nothingaccept_load0start0odata0round1def
, accept_load0start0odata1def
);
signal cmd : CONTROL;


begin


   --portmap
   label_CORE : CORE port map (
         message => message_reg,
         Hashvalue => s_odata,
         start => start,
         EoM => EoM,
         shift => r_shift,
         rst_n => rst_n,
         clk => clk
      );
   --register updates
   dpREG: process (clk,rst_n)
      begin
         if (rst_n = '0') then
            load_reg <= '0';
            fetch_reg <= '0';
            init_reg <= '0';
            ack_reg <= '0';
            message_reg <= (others=>'0');
            round <= (others=>'0');
         elsif clk' event and clk = '1' then
            load_reg <= load_reg_wire;
            fetch_reg <= fetch_reg_wire;
            init_reg <= init_reg_wire;
            ack_reg <= ack_reg_wire;
            message_reg <= message_reg_wire;
            round <= round_wire;

         end if;
      end process dpREG;


   --combinational logics
   dpCMB: process (load_reg,fetch_reg,init_reg,load_ack,fetch_ack,s_idata,s_odata,accept_load,odata_ready,ack_reg
,message_reg,r_shift,start,round,sig_0,sig_1,sig_2,sig_3,sig_4,sig_5
,sig_6,sig_7,odata_int,ack_int,init,EoM,load,fetch,idata,cmd,STATE)
      begin
         load_reg_wire <= load_reg;
         fetch_reg_wire <= fetch_reg;
         init_reg_wire <= init_reg;
         load_ack <= '0';
         fetch_ack <= '0';
         s_idata <= (others=>'0');
         accept_load <= '0';
         odata_ready <= '0';
         ack_reg_wire <= ack_reg;
         message_reg_wire <= message_reg;
         start <= '0';
         round_wire <= round;
         sig_0 <= '0';
         sig_1 <= '0';
         sig_2 <= '0';
         sig_3 <= '0';
         sig_4 <= '0';
         sig_5 <= '0';
         sig_6 <= (others=>'0');
         sig_7 <= (others=>'0');
         odata_int <= (others=>'0');
         ack_int <= '0';
         ack <= '0';
         odata <= (others=>'0');



         case cmd is
            when initializedef =>
               if (r_shift = '1') then
                  sig_5 <= '1';
               else
                  sig_5 <= '0';
               end if;
               accept_load <= sig_5;
               odata_ready <= '0';
               start <= '0';
               message_reg_wire <= conv_std_logic_vector(unsigned(s_idata),512);
               round_wire <= conv_std_logic_vector(1,5);
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when nothingstart0odata0accept_load0def =>
               start <= '0';
               odata_ready <= '0';
               accept_load <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when start0accept_load1odata0def =>
               start <= '0';
               accept_load <= '1';
               odata_ready <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when take_in_odataaccept_load0start0odata0def =>
               sig_6 <= message_reg(495 downto 0) & s_idata;
               message_reg_wire <= sig_6;
               accept_load <= '0';
               start <= '0';
               odata_ready <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when start1accept_load0odata0def =>
               start <= '1';
               accept_load <= '0';
               odata_ready <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when round0start0accept_load1odata0def =>
               round_wire <= conv_std_logic_vector(0,5);
               start <= '0';
               accept_load <= '1';
               odata_ready <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when round1start0accept_load1odata0def =>
               sig_7 <= unsigned(round) + unsigned(conv_std_logic_vector(1,5));
               round_wire <= sig_7;
               start <= '0';
               accept_load <= '1';
               odata_ready <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when accept_load0start0odata0def =>
               accept_load <= '0';
               start <= '0';
               odata_ready <= '0';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when nothingaccept_load0start0odata0round1def =>
               accept_load <= '0';
               start <= '0';
               odata_ready <= '0';
               sig_7 <= unsigned(round) + unsigned(conv_std_logic_vector(1,5));
               round_wire <= sig_7;
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when accept_load0start0odata1def =>
               accept_load <= '0';
               start <= '0';
               odata_ready <= '1';
               odata <= odata_int;
               s_idata <= idata;
               sig_0 <= fetch and odata_ready;
               fetch_ack <= sig_0;
               sig_1 <=  not fetch_reg;
               sig_2 <= load_reg and accept_load;
               if (sig_1 = '1') then
                  sig_3 <= sig_2;
               else
                  sig_3 <= '0';
               end if;
               load_ack <= sig_3;
               sig_4 <= fetch_ack or load_ack;
               ack <= ack_int;
               odata_int <= s_odata;
               ack_int <= sig_4;
               fetch_reg_wire <= fetch;
               init_reg_wire <= init;
               load_reg_wire <= load;
            when others=>
         end case;
      end process dpCMB;


   --controller reg
   fsmREG: process (clk,rst_n)
      begin
         if (rst_n = '0') then
            STATE <= S_uninit;
         elsif clk' event and clk = '1' then
            STATE <= STATE;
            case STATE is
               when S_uninit => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                          STATE <= S_uninit;
                  end if;
               when S_ready => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                     if (load_reg = '1') then
                             STATE <= S_loadstep1;
                     else
                        if (fetch_reg = '1') then
                                STATE <= S_fetch;
                        else
                                STATE <= S_ready;
                        end if;
                     end if;
                  end if;
               when S_loading => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                     if (sig_9 = '1') then
                             STATE <= S_calc;
                     else
                        if (load_reg = '1') then
                                STATE <= S_loadstep1;
                        else
                                STATE <= S_loading;
                        end if;
                     end if;
                  end if;
               when S_calc => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                     if (r_shift = '1') then
                             STATE <= S_ready;
                     else
                             STATE <= S_calc;
                     end if;
                  end if;
               when S_fetch => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                     if (sig_10 = '1') then
                             STATE <= S_fetch;
                     else
                        if (sig_11 = '1') then
                                STATE <= S_ready;
                        else
                           if (fetch_reg = '1') then
                                   STATE <= S_fetchstep1;
                           else
                                   STATE <= S_fetch;
                           end if;
                        end if;
                     end if;
                  end if;
               when S_loadstep1 => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                          STATE <= S_loadstep2;
                  end if;
               when S_loadstep2 => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                     if (sig_8 = '1') then
                             STATE <= S_loading;
                     else
                             STATE <= S_loading;
                     end if;
                  end if;
               when S_fetchstep1 => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                          STATE <= S_fetchstep2;
                  end if;
               when S_fetchstep2 => 
                  if (init_reg = '1') then
                          STATE <= S_ready;
                  else
                          STATE <= S_fetch;
                  end if;
               when others=>
            end case;
         end if;
      end process fsmREG;


   --controller cmb
   fsmCMB: process (load_reg,fetch_reg,init_reg,load_ack,fetch_ack,s_idata,s_odata,accept_load,odata_ready,ack_reg
,message_reg,r_shift,start,round,sig_0,sig_1,sig_2,sig_3,sig_4,sig_5
,sig_6,sig_7,odata_int,ack_int,sig_8,sig_9,sig_10,sig_11,init,EoM
,load,fetch,idata,cmd,STATE)
      begin
      sig_8 <= '0';
      sig_9 <= '0';
      sig_10 <= '0';
      sig_11 <= '0';
      if (unsigned(round) = 31) then
         sig_8 <= '1';
      else
         sig_8 <= '0';
      end if;
      if (unsigned(round) = 31) then
         sig_9 <= '1';
      else
         sig_9 <= '0';
      end if;
      if (unsigned(round) = 0) then
         sig_10 <= '1';
      else
         sig_10 <= '0';
      end if;
      if (unsigned(round) = 17) then
         sig_11 <= '1';
      else
         sig_11 <= '0';
      end if;
      cmd <= initializedef;
      case STATE is
         when S_uninit => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
                    cmd <= nothingstart0odata0accept_load0def;
            end if;
         when S_ready => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
               if (load_reg = '1') then
                       cmd <= start0accept_load1odata0def;
               else
                  if (fetch_reg = '1') then
                          cmd <= nothingstart0odata0accept_load0def;
                  else
                          cmd <= initializedef;
                  end if;
               end if;
            end if;
         when S_loading => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
               if (sig_9 = '1') then
                       cmd <= round0start0accept_load1odata0def;
               else
                  if (load_reg = '1') then
                          cmd <= round1start0accept_load1odata0def;
                  else
                          cmd <= nothingstart0odata0accept_load0def;
                  end if;
               end if;
            end if;
         when S_calc => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
               if (r_shift = '1') then
                       cmd <= accept_load0start0odata0def;
               else
                       cmd <= nothingstart0odata0accept_load0def;
               end if;
            end if;
         when S_fetch => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
               if (sig_10 = '1') then
                       cmd <= nothingaccept_load0start0odata0round1def;
               else
                  if (sig_11 = '1') then
                          cmd <= accept_load0start0odata0def;
                  else
                     if (fetch_reg = '1') then
                             cmd <= accept_load0start0odata1def;
                     else
                             cmd <= nothingstart0odata0accept_load0def;
                     end if;
                  end if;
               end if;
            end if;
         when S_loadstep1 => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
                    cmd <= take_in_odataaccept_load0start0odata0def;
            end if;
         when S_loadstep2 => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
               if (sig_8 = '1') then
                       cmd <= start1accept_load0odata0def;
               else
                       cmd <= nothingstart0odata0accept_load0def;
               end if;
            end if;
         when S_fetchstep1 => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
                    cmd <= nothingstart0odata0accept_load0def;
            end if;
         when S_fetchstep2 => 
            if (init_reg = '1') then
                    cmd <= initializedef;
            else
                    cmd <= nothingaccept_load0start0odata0round1def;
            end if;
         when others=>
         end case;
   end process fsmCMB;
end RTL;
